=== modified file 'NEWS'
--- NEWS	2006-07-17 19:02:45 +0000
+++ NEWS	2006-07-19 18:02:26 +0000
@@ -104,6 +104,14 @@
    * 'bzr init' now also works on remote locations.
      (Wouter van Heyst, #48904)
 
+   * HTTP support has been updated. When using pycurl we now support 
+     connection keep-alive, which reduces dns requests and round trips.
+     And for both urllib and pycurl we support multi-range requests, 
+     which decreases the number of round-trips. Performance results for
+     ``bzr branch http://bazaar-vcs.org/bzr/bzr.dev/`` is not 2-3x faster.
+     And ``bzr pull`` in an existing branch is as much as 4x faster.
+     (Michael Ellerman, Johan Rydberg, John Arbash Meinel, #46768)
+
   BUG FIXES:
 
     * Fix shadowed definition of TestLocationConfig that caused some 

=== modified file 'bzrlib/errors.py'
--- bzrlib/errors.py	2006-07-12 02:20:47 +0000
+++ bzrlib/errors.py	2006-07-13 18:43:34 +0000
@@ -749,6 +749,38 @@
     """Connection closed: %(msg)s %(orig_error)s"""
 
 
+class InvalidRange(TransportError):
+    """Invalid range access."""
+    
+    def __init__(self, path, offset):
+        TransportError.__init__(self, ("Invalid range access in %s at %d"
+                                       % (path, offset)))
+
+
+class InvalidHttpResponse(TransportError):
+    """Invalid http response for %(path)s: %(msg)s"""
+
+    def __init__(self, path, msg, orig_error=None):
+        self.path = path
+        TransportError.__init__(self, msg, orig_error=orig_error)
+
+
+class InvalidHttpRange(InvalidHttpResponse):
+    """Invalid http range "%(range)s" for %(path)s: %(msg)s"""
+    
+    def __init__(self, path, range, msg):
+        self.range = range
+        InvalidHttpResponse.__init__(self, path, msg)
+
+
+class InvalidHttpContentType(InvalidHttpResponse):
+    """Invalid http Content-type "%(ctype)s" for %(path)s: %(msg)s"""
+    
+    def __init__(self, path, ctype, msg):
+        self.ctype = ctype
+        InvalidHttpResponse.__init__(self, path, msg)
+
+
 class ConflictsInTree(BzrError):
     def __init__(self):
         BzrError.__init__(self, "Working tree has conflicts.")

=== modified file 'bzrlib/tests/__init__.py'
--- bzrlib/tests/__init__.py	2006-06-30 15:13:19 +0000
+++ bzrlib/tests/__init__.py	2006-07-06 16:29:12 +0000
@@ -1263,6 +1263,7 @@
                    'bzrlib.tests.test_graph',
                    'bzrlib.tests.test_hashcache',
                    'bzrlib.tests.test_http',
+                   'bzrlib.tests.test_http_response',
                    'bzrlib.tests.test_identitymap',
                    'bzrlib.tests.test_inv',
                    'bzrlib.tests.test_knit',

=== modified file 'bzrlib/tests/test_http.py'
--- bzrlib/tests/test_http.py	2006-05-06 17:54:19 +0000
+++ bzrlib/tests/test_http.py	2006-07-14 19:25:24 +0000
@@ -23,11 +23,13 @@
 from bzrlib.errors import DependencyNotPresent
 from bzrlib.tests import TestCase, TestSkipped
 from bzrlib.transport import Transport
-from bzrlib.transport.http import extract_auth
+from bzrlib.transport.http import extract_auth, HttpTransportBase
 from bzrlib.transport.http._urllib import HttpTransport_urllib
 from bzrlib.tests.HTTPTestUtil import TestCaseWithWebserver
 
+
 class FakeManager (object):
+
     def __init__(self):
         self.credentials = []
         
@@ -36,6 +38,7 @@
 
 
 class TestHttpUrls(TestCase):
+
     def test_url_parsing(self):
         f = FakeManager()
         url = extract_auth('http://example.com', f)
@@ -90,6 +93,7 @@
         except DependencyNotPresent:
             raise TestSkipped('pycurl not present')
 
+
 class TestHttpMixins(object):
 
     def _prep_tree(self):
@@ -124,6 +128,7 @@
 
 
 class TestHttpConnections_urllib(TestCaseWithWebserver, TestHttpMixins):
+
     _transport = HttpTransport_urllib
 
     def setUp(self):
@@ -159,3 +164,52 @@
         t = get_transport('http+urllib://bzr.google.com/')
         self.assertIsInstance(t, Transport)
         self.assertIsInstance(t, bzrlib.transport.http._urllib.HttpTransport_urllib)
+
+
+class TestOffsets(TestCase):
+    """Test offsets_to_ranges method"""
+
+    def test_offsets_to_ranges_simple(self):
+        to_range = HttpTransportBase.offsets_to_ranges
+        ranges = to_range([(10, 1)])
+        self.assertEqual([[10, 10]], ranges)
+
+        ranges = to_range([(0, 1), (1, 1)])
+        self.assertEqual([[0, 1]], ranges)
+
+        ranges = to_range([(1, 1), (0, 1)])
+        self.assertEqual([[0, 1]], ranges)
+
+    def test_offset_to_ranges_overlapped(self):
+        to_range = HttpTransportBase.offsets_to_ranges
+
+        ranges = to_range([(10, 1), (20, 2), (22, 5)])
+        self.assertEqual([[10, 10], [20, 26]], ranges)
+
+        ranges = to_range([(10, 1), (11, 2), (22, 5)])
+        self.assertEqual([[10, 12], [22, 26]], ranges)
+
+
+class TestRangeHeader(TestCase):
+    """Test range_header method"""
+
+    def check_header(self, value, ranges=[], tail=0):
+        range_header = HttpTransportBase.range_header
+        self.assertEqual(value, range_header(ranges, tail))
+
+    def test_range_header_single(self):
+        self.check_header('0-9', ranges=[[0,9]])
+        self.check_header('100-109', ranges=[[100,109]])
+
+    def test_range_header_tail(self):
+        self.check_header('-10', tail=10)
+        self.check_header('-50', tail=50)
+
+    def test_range_header_multi(self):
+        self.check_header('0-9,100-200,300-5000',
+                          ranges=[(0,9), (100, 200), (300,5000)])
+
+    def test_range_header_mixed(self):
+        self.check_header('0-9,300-5000,-50',
+                          ranges=[(0,9), (300,5000)],
+                          tail=50)

=== added file 'bzrlib/tests/test_http_response.py'
--- bzrlib/tests/test_http_response.py	1970-01-01 00:00:00 +0000
+++ bzrlib/tests/test_http_response.py	2006-07-19 14:17:49 +0000
@@ -0,0 +1,600 @@
+# Copyright (C) 2005, 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests from HTTP response parsing."""
+
+from cStringIO import StringIO
+import mimetools
+
+from bzrlib import errors
+from bzrlib.transport import http
+from bzrlib.transport.http import response
+from bzrlib.tests import TestCase
+
+
+class TestResponseRange(TestCase):
+    """Test the ResponseRange class."""
+
+    def test_cmp(self):
+        RR = response.ResponseRange
+        r1 = RR(0, 10, 0)
+        r2 = RR(15, 20, 10)
+        self.assertTrue(r1 < r2)
+        self.assertFalse(r1 > r2)
+        self.assertTrue(r1 < 5)
+        self.assertFalse(r2 < 5)
+
+        self.assertEqual(RR(0, 10, 5), RR(0, 10, 5))
+        self.assertNotEqual(RR(0, 10, 5), RR(0, 8, 5))
+        self.assertNotEqual(RR(0, 10, 5), RR(0, 10, 6))
+
+    def test_sort_list(self):
+        """Ensure longer ranges are sorted after shorter ones"""
+        RR = response.ResponseRange
+        lst = [RR(3, 8, 0), 5, RR(3, 7, 0), 6]
+        lst.sort()
+        self.assertEqual([RR(3,7,0), RR(3,8,0), 5, 6], lst)
+
+
+class TestRangeFile(TestCase):
+    """Test RangeFile."""
+
+    def setUp(self):
+        content = "abcdefghijklmnopqrstuvwxyz"
+        self.fp = response.RangeFile('foo', StringIO(content))
+        self.fp._add_range(0,  9,   0)
+        self.fp._add_range(20, 29, 10)
+        self.fp._add_range(30, 39, 15)
+
+    def test_valid_accesses(self):
+        """Test so that valid accesses work to the file."""
+        self.fp.seek(0, 0)
+        self.assertEquals(self.fp.read(3), 'abc')
+        self.assertEquals(self.fp.read(3), 'def')
+        self.assertEquals(self.fp.tell(), 6)
+        self.fp.seek(20, 0)
+        self.assertEquals(self.fp.read(3), 'klm')
+        self.assertEquals(self.fp.read(2), 'no')
+        self.assertEquals(self.fp.tell(), 25)
+        # should wrap over to 30-39 entity
+        self.assertEquals(self.fp.read(3), 'pqr')
+        self.fp.seek(3)
+        self.assertEquals(self.fp.read(3), 'def')
+        self.assertEquals(self.fp.tell(), 6)
+
+    def test_invalid_accesses(self):
+        """Test so that invalid accesses trigger errors."""
+        self.fp.seek(9)
+        self.assertRaises(errors.InvalidRange, self.fp.read, 2)
+        self.fp.seek(39)
+        self.assertRaises(errors.InvalidRange, self.fp.read, 2)
+        self.fp.seek(19)
+        self.assertRaises(errors.InvalidRange, self.fp.read, 2)
+
+    def test__finish_ranges(self):
+        """Test that after RangeFile._finish_ranges the list is sorted."""
+        self.fp._add_range(1, 2, 3)
+        self.fp._add_range(8, 9, 10)
+        self.fp._add_range(3, 4, 5)
+
+        # TODO: jam 20060706 If we switch to inserting
+        #       in sorted order, remove this test
+        self.assertNotEqual(self.fp._ranges, sorted(self.fp._ranges))
+
+        self.fp._finish_ranges()
+        self.assertEqual(self.fp._ranges, sorted(self.fp._ranges))
+
+    def test_seek_and_tell(self):
+        # Check for seeking before start
+        self.fp.seek(-2, 0)
+        self.assertEqual(0, self.fp.tell())
+
+        self.fp.seek(5, 0)
+        self.assertEqual(5, self.fp.tell())
+
+        self.fp.seek(-2, 1)
+        self.assertEqual(3, self.fp.tell())
+
+        # TODO: jam 20060706 following tests will fail if this 
+        #       is not true, and would be difficult to debug
+        #       but it is a layering violation
+        self.assertEqual(39, self.fp._len)
+
+        self.fp.seek(0, 2)
+        self.assertEqual(39, self.fp.tell())
+
+        self.fp.seek(-10, 2)
+        self.assertEqual(29, self.fp.tell())
+
+        self.assertRaises(ValueError, self.fp.seek, 0, 4)
+        self.assertRaises(ValueError, self.fp.seek, 0, -1)
+
+
+class TestRegexes(TestCase):
+
+    def assertRegexMatches(self, groups, text):
+        """Check that the regex matches and returns the right values"""
+        m = self.regex.match(text)
+        self.assertNotEqual(None, m, "text %s did not match regex" % (text,))
+
+        self.assertEqual(groups, m.groups())
+
+    def test_range_re(self):
+        """Test that we match valid ranges."""
+        self.regex = response.HttpRangeResponse._CONTENT_RANGE_RE
+        self.assertRegexMatches(('bytes', '1', '10', '11'),
+                           'bytes 1-10/11')
+        self.assertRegexMatches(('bytes', '1', '10', '11'),
+                           '\tbytes  1-10/11   ')
+        self.assertRegexMatches(('bytes', '2123', '4242', '1231'),
+                           '\tbytes  2123-4242/1231   ')
+        self.assertRegexMatches(('chars', '1', '2', '3'),
+                           ' chars 1-2/3')
+
+    def test_content_type_re(self):
+        self.regex = response.HttpMultipartRangeResponse._CONTENT_TYPE_RE
+        self.assertRegexMatches(('xxyyzz',),
+                                'multipart/byteranges; boundary = xxyyzz')
+        self.assertRegexMatches(('xxyyzz',),
+                                'multipart/byteranges;boundary=xxyyzz')
+        self.assertRegexMatches(('xx yy zz',),
+                                ' multipart/byteranges ; boundary= xx yy zz ')
+        self.assertEqual(None,
+                self.regex.match('multipart byteranges;boundary=xx'))
+
+
+simple_data = """
+--xxyyzz\r
+foo\r
+Content-range: bytes 1-10/20\r
+\r
+1234567890
+--xxyyzz\r
+Content-Range: bytes 21-30/20\r
+bar\r
+\r
+abcdefghij
+--xxyyzz\r
+content-range: bytes 41-50/20\r
+\r
+zyxwvutsrq
+--xxyyzz\r
+content-range: bytes 51-60/20\r
+\r
+xxyyzz fbd
+"""
+
+
+class TestHelpers(TestCase):
+    """Test the helper functions"""
+
+    def test__parse_range(self):
+        """Test that _parse_range acts reasonably."""
+        content = StringIO('')
+        parse_range = response.HttpRangeResponse._parse_range
+        self.assertEqual((1,2), parse_range('bytes 1-2/3'))
+        self.assertEqual((10,20), parse_range('bytes 10-20/2'))
+
+        self.assertRaises(errors.InvalidHttpRange, parse_range, 'char 1-3/2')
+        self.assertRaises(errors.InvalidHttpRange, parse_range, 'bytes a-3/2')
+
+        try:
+            parse_range('bytes x-10/3', path='http://foo/bar')
+        except errors.InvalidHttpRange, e:
+            self.assertContainsRe(str(e), 'http://foo/bar')
+            self.assertContainsRe(str(e), 'bytes x-10/3')
+        else:
+            self.fail('Did not raise InvalidHttpRange')
+
+    def test__parse_boundary_simple(self):
+        """Test that _parse_boundary handles Content-type properly"""
+        parse_boundary = response.HttpMultipartRangeResponse._parse_boundary
+        m = parse_boundary(' multipart/byteranges; boundary=xxyyzz')
+        self.assertNotEqual(None, m)
+        # Check that the returned regex is capable of splitting simple_data
+        matches = list(m.finditer(simple_data))
+        self.assertEqual(4, len(matches))
+
+        # match.group() should be the content-range entry
+        # and match.end() should be the start of the content
+        self.assertEqual(' bytes 1-10/20', matches[0].group(1))
+        self.assertEqual(simple_data.find('1234567890'), matches[0].end())
+        self.assertEqual(' bytes 21-30/20', matches[1].group(1))
+        self.assertEqual(simple_data.find('abcdefghij'), matches[1].end())
+        self.assertEqual(' bytes 41-50/20', matches[2].group(1))
+        self.assertEqual(simple_data.find('zyxwvutsrq'), matches[2].end())
+        self.assertEqual(' bytes 51-60/20', matches[3].group(1))
+        self.assertEqual(simple_data.find('xxyyzz fbd'), matches[3].end())
+
+    def test__parse_boundary_invalid(self):
+        parse_boundary = response.HttpMultipartRangeResponse._parse_boundary
+        try:
+            parse_boundary(' multipart/bytes;boundary=xxyyzz',
+                           path='http://foo/bar')
+        except errors.InvalidHttpContentType, e:
+            self.assertContainsRe(str(e), 'http://foo/bar')
+            self.assertContainsRe(str(e), 'multipart/bytes;boundary=xxyyzz')
+        else:
+            self.fail('Did not raise InvalidHttpContentType')
+
+
+class TestHttpRangeResponse(TestCase):
+
+    def test_smoketest(self):
+        """A basic test that HttpRangeResponse is reasonable."""
+        content = StringIO('0123456789')
+        f = response.HttpRangeResponse('http://foo', 'bytes 1-10/9', content)
+        self.assertEqual([response.ResponseRange(1,10,0)], f._ranges)
+
+        f.seek(0)
+        self.assertRaises(errors.InvalidRange, f.read, 2)
+        f.seek(1)
+        self.assertEqual('012345', f.read(6))
+
+    def test_invalid(self):
+        try:
+            f = response.HttpRangeResponse('http://foo', 'bytes x-10/9',
+                                           StringIO('0123456789'))
+        except errors.InvalidHttpRange, e:
+            self.assertContainsRe(str(e), 'http://foo')
+            self.assertContainsRe(str(e), 'bytes x-10/9')
+        else:
+            self.fail('Failed to raise InvalidHttpRange')
+
+
+class TestHttpMultipartRangeResponse(TestCase):
+    """Test the handling of multipart range responses"""
+
+    def test_simple(self):
+        content = StringIO(simple_data)
+        multi = response.HttpMultipartRangeResponse('http://foo',
+                    'multipart/byteranges; boundary = xxyyzz', content)
+
+        self.assertEqual(4, len(multi._ranges))
+
+        multi.seek(1)
+        self.assertEqual('1234567890', multi.read(10))
+        multi.seek(21)
+        self.assertEqual('abcdefghij', multi.read(10))
+        multi.seek(41)
+        self.assertEqual('zyxwvutsrq', multi.read(10))
+        multi.seek(51)
+        self.assertEqual('xxyyzz fbd', multi.read(10))
+        # TODO: jam 20060706 Currently RangeFile does not support
+        #       reading across ranges. Consider adding it.
+        multi.seek(41)
+        # self.assertEqual('zyxwvutsrqxxyyzz fbd', multi.read(20))
+        self.assertRaises(errors.InvalidRange, multi.read, 20)
+
+        multi.seek(21)
+        self.assertRaises(errors.InvalidRange, multi.read, 11)
+        multi.seek(31)
+        self.assertRaises(errors.InvalidRange, multi.read, 10)
+
+    def test_invalid(self):
+        content = StringIO('')
+        try:
+            response.HttpMultipartRangeResponse('http://foo',
+                        'multipart/byte;boundary=invalid', content)
+        except errors.InvalidHttpContentType, e:
+            self.assertContainsRe(str(e), 'http://foo')
+            self.assertContainsRe(str(e), 'multipart/byte;')
+
+
+# Taken from real request responses
+_full_text_response = (200, """HTTP/1.1 200 OK\r
+Date: Tue, 11 Jul 2006 04:32:56 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Sun, 23 Apr 2006 19:35:20 GMT\r
+ETag: "56691-23-38e9ae00"\r
+Accept-Ranges: bytes\r
+Content-Length: 35\r
+Connection: close\r
+Content-Type: text/plain; charset=UTF-8\r
+\r
+""", """Bazaar-NG meta directory, format 1
+""")
+
+
+_missing_response = (404, """HTTP/1.1 404 Not Found\r
+Date: Tue, 11 Jul 2006 04:32:56 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Content-Length: 336\r
+Connection: close\r
+Content-Type: text/html; charset=iso-8859-1\r
+\r
+""", """<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
+<html><head>
+<title>404 Not Found</title>
+</head><body>
+<h1>Not Found</h1>
+<p>The requested URL /branches/bzr/jam-integration/.bzr/repository/format was not found on this server.</p>
+<hr>
+<address>Apache/2.0.54 (Fedora) Server at bzr.arbash-meinel.com Port 80</address>
+</body></html>
+""")
+
+
+_single_range_response = (206, """HTTP/1.1 206 Partial Content\r
+Date: Tue, 11 Jul 2006 04:45:22 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Thu, 06 Jul 2006 20:22:05 GMT\r
+ETag: "238a3c-16ec2-805c5540"\r
+Accept-Ranges: bytes\r
+Content-Length: 100\r
+Content-Range: bytes 100-199/93890\r
+Connection: close\r
+Content-Type: text/plain; charset=UTF-8\r
+\r
+""", """mbp@sourcefrog.net-20050309040815-13242001617e4a06
+mbp@sourcefrog.net-20050309040929-eee0eb3e6d1e762""")
+
+
+_multipart_range_response = (206, """HTTP/1.1 206 Partial Content\r
+Date: Tue, 11 Jul 2006 04:49:48 GMT\r
+Server: Apache/2.0.54 (Fedora)\r
+Last-Modified: Thu, 06 Jul 2006 20:22:05 GMT\r
+ETag: "238a3c-16ec2-805c5540"\r
+Accept-Ranges: bytes\r
+Content-Length: 1534\r
+Connection: close\r
+Content-Type: multipart/byteranges; boundary=418470f848b63279b\r
+\r
+\r""", """--418470f848b63279b\r
+Content-type: text/plain; charset=UTF-8\r
+Content-range: bytes 0-254/93890\r
+\r
+mbp@sourcefrog.net-20050309040815-13242001617e4a06
+mbp@sourcefrog.net-20050309040929-eee0eb3e6d1e7627
+mbp@sourcefrog.net-20050309040957-6cad07f466bb0bb8
+mbp@sourcefrog.net-20050309041501-c840e09071de3b67
+mbp@sourcefrog.net-20050309044615-c24a3250be83220a
+\r
+--418470f848b63279b\r
+Content-type: text/plain; charset=UTF-8\r
+Content-range: bytes 1000-2049/93890\r
+\r
+40-fd4ec249b6b139ab
+mbp@sourcefrog.net-20050311063625-07858525021f270b
+mbp@sourcefrog.net-20050311231934-aa3776aff5200bb9
+mbp@sourcefrog.net-20050311231953-73aeb3a131c3699a
+mbp@sourcefrog.net-20050311232353-f5e33da490872c6a
+mbp@sourcefrog.net-20050312071639-0a8f59a34a024ff0
+mbp@sourcefrog.net-20050312073432-b2c16a55e0d6e9fb
+mbp@sourcefrog.net-20050312073831-a47c3335ece1920f
+mbp@sourcefrog.net-20050312085412-13373aa129ccbad3
+mbp@sourcefrog.net-20050313052251-2bf004cb96b39933
+mbp@sourcefrog.net-20050313052856-3edd84094687cb11
+mbp@sourcefrog.net-20050313053233-e30a4f28aef48f9d
+mbp@sourcefrog.net-20050313053853-7c64085594ff3072
+mbp@sourcefrog.net-20050313054757-a86c3f5871069e22
+mbp@sourcefrog.net-20050313061422-418f1f73b94879b9
+mbp@sourcefrog.net-20050313120651-497bd231b19df600
+mbp@sourcefrog.net-20050314024931-eae0170ef25a5d1a
+mbp@sourcefrog.net-20050314025438-d52099f915fe65fc
+mbp@sourcefrog.net-20050314025539-637a636692c055cf
+mbp@sourcefrog.net-20050314025737-55eb441f430ab4ba
+mbp@sourcefrog.net-20050314025901-d74aa93bb7ee8f62
+mbp@source\r
+--418470f848b63279b--\r\n'
+""")
+
+
+_redirect_response = (206, """HTTP/1.1 301 Moved Permanently\r
+Date: Tue, 18 Jul 2006 20:29:22 GMT\r
+Server: Apache/2.0.54 (Ubuntu) PHP/4.4.0-3ubuntu1 mod_ssl/2.0.54 OpenSSL/0.9.7g\r
+Location: http://bazaar-vcs.org/bzr/bzr.dev/.bzr/repository/inventory.knit\r
+Content-Length: 272\r
+Keep-Alive: timeout=15, max=100\r
+Connection: Keep-Alive\r
+Content-Type: text/html; charset=iso-8859-1\r
+\r
+HTTP/1.1 206 Partial Content\r
+Date: Tue, 18 Jul 2006 20:29:23 GMT\r
+Server: Apache/2.0.54 (Ubuntu) PHP/4.4.0-3ubuntu1 mod_ssl/2.0.54 OpenSSL/0.9.7g\r
+Last-Modified: Tue, 18 Jul 2006 20:24:59 GMT\r
+ETag: "be8213-83958c-f0d3dcc0"\r
+Accept-Ranges: bytes\r
+Content-Length: 425\r
+Content-Range: bytes 8623075-8623499/8623500\r
+Keep-Alive: timeout=15, max=100\r
+Connection: Keep-Alive\r
+Content-Type: text/plain; charset=UTF-8\r
+\r
+""", """this data intentionally removed, 
+this is not meant to be tested by
+handle_response, just _extract_headers
+""")
+
+
+# This is made up
+_invalid_response = (444, """HTTP/1.1 444 Bad Response\r
+Date: Tue, 11 Jul 2006 04:32:56 GMT\r
+Connection: close\r
+Content-Type: text/html; charset=iso-8859-1\r
+\r
+""", """<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
+<html><head>
+<title>404 Not Found</title>
+</head><body>
+<h1>Not Found</h1>
+<p>I don't know what I'm doing</p>
+<hr>
+</body></html>
+""")
+
+
+# This should be in test_http.py, but the headers we
+# want to parse are here
+class TestExtractHeader(TestCase):
+    
+    def use_response(self, response):
+        self.headers = http._extract_headers(response[1], 'http://foo')
+
+    def check_header(self, header, value):
+        self.assertEqual(value, self.headers[header])
+        
+    def test_full_text(self):
+        self.use_response(_full_text_response)
+
+        self.check_header('Date', 'Tue, 11 Jul 2006 04:32:56 GMT')
+        self.check_header('date', 'Tue, 11 Jul 2006 04:32:56 GMT')
+        self.check_header('Content-Length', '35')
+        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
+        self.check_header('content-type', 'text/plain; charset=UTF-8')
+
+    def test_missing_response(self):
+        self.use_response(_missing_response)
+
+        self.check_header('Content-Length', '336')
+        self.check_header('Content-Type', 'text/html; charset=iso-8859-1')
+
+    def test_single_range(self):
+        self.use_response(_single_range_response)
+
+        self.check_header('Content-Length', '100')
+        self.check_header('Content-Range', 'bytes 100-199/93890')
+        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
+
+    def test_multi_range(self):
+        self.use_response(_multipart_range_response)
+
+        self.check_header('Content-Length', '1534')
+        self.check_header('Content-Type',
+                          'multipart/byteranges; boundary=418470f848b63279b')
+
+    def test_redirect(self):
+        """We default to returning the last group of headers in the file."""
+        self.use_response(_redirect_response)
+        self.check_header('Content-Range', 'bytes 8623075-8623499/8623500')
+        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
+
+    def test_empty(self):
+        self.assertRaises(errors.InvalidHttpResponse,
+            http._extract_headers, '', 'bad url')
+
+    def test_no_opening_http(self):
+        # Remove the HTTP line from the header
+        first, txt = _full_text_response[1].split('\r\n', 1)
+        self.assertRaises(errors.InvalidHttpResponse,
+            http._extract_headers, txt, 'missing HTTTP')
+
+    def test_trailing_whitespace(self):
+        # Test that we ignore bogus whitespace on the end
+        code, txt, body = _full_text_response
+        txt += '\r\n\n\n\n\n'
+        self.use_response((code, txt, body))
+
+        self.check_header('Date', 'Tue, 11 Jul 2006 04:32:56 GMT')
+        self.check_header('Content-Length', '35')
+        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
+
+    def test_trailing_non_http(self):
+        # Test that we ignore bogus stuff on the end
+        code, txt, body = _full_text_response
+        txt = txt + 'Foo: Bar\r\nBaz: Bling\r\n\r\n'
+        self.use_response((code, txt, body))
+
+        self.check_header('Date', 'Tue, 11 Jul 2006 04:32:56 GMT')
+        self.check_header('Content-Length', '35')
+        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
+        self.assertRaises(KeyError, self.headers.__getitem__, 'Foo')
+
+    def test_extra_whitespace(self):
+        # Test that we read an HTTP response, even with extra whitespace
+        code, txt, body = _redirect_response
+        # Find the second HTTP location
+        loc = txt.find('HTTP', 5)
+        txt = txt[:loc] + '\r\n\n' + txt[loc:]
+        self.use_response((code, txt, body))
+        self.check_header('Content-Range', 'bytes 8623075-8623499/8623500')
+        self.check_header('Content-Type', 'text/plain; charset=UTF-8')
+
+
+class TestHandleResponse(TestCase):
+    
+    def get_response(self, a_response):
+        """Process a supplied response, and return the result."""
+        headers = http._extract_headers(a_response[1], 'http://foo')
+        return response.handle_response('http://foo', a_response[0], headers,
+                                        StringIO(a_response[2]))
+
+    def test_full_text(self):
+        out = self.get_response(_full_text_response)
+        # It is a StringIO from the original data
+        self.assertEqual(_full_text_response[2], out.read())
+
+    def test_missing_response(self):
+        self.assertRaises(errors.NoSuchFile,
+            self.get_response, _missing_response)
+
+    def test_single_range(self):
+        out = self.get_response(_single_range_response)
+        self.assertIsInstance(out, response.HttpRangeResponse)
+
+        self.assertRaises(errors.InvalidRange, out.read, 20)
+
+        out.seek(100)
+        self.assertEqual(_single_range_response[2], out.read(100))
+
+    def test_multi_range(self):
+        out = self.get_response(_multipart_range_response)
+        self.assertIsInstance(out, response.HttpMultipartRangeResponse)
+
+        # Just make sure we can read the right contents
+        out.seek(0)
+        out.read(255)
+
+        out.seek(1000)
+        out.read(1050)
+
+    def test_invalid_response(self):
+        self.assertRaises(errors.InvalidHttpResponse,
+            self.get_response, _invalid_response)
+
+    def test_full_text_no_content_type(self):
+        # We should not require Content-Type for a full response
+        a_response = _full_text_response
+        headers = http._extract_headers(a_response[1], 'http://foo')
+        del headers['Content-Type']
+        out = response.handle_response('http://foo', a_response[0], headers,
+                                        StringIO(a_response[2]))
+        self.assertEqual(_full_text_response[2], out.read())
+
+    def test_missing_no_content_type(self):
+        # Without Content-Type we should still raise NoSuchFile on a 404
+        a_response = _missing_response
+        headers = http._extract_headers(a_response[1], 'http://missing')
+        del headers['Content-Type']
+        self.assertRaises(errors.NoSuchFile,
+            response.handle_response, 'http://missing', a_response[0], headers,
+                                      StringIO(a_response[2]))
+
+    def test_missing_content_type(self):
+        a_response = _single_range_response
+        headers = http._extract_headers(a_response[1], 'http://nocontent')
+        del headers['Content-Type']
+        self.assertRaises(errors.InvalidHttpContentType,
+            response.handle_response, 'http://nocontent', a_response[0],
+                                      headers, StringIO(a_response[2]))
+
+    def test_missing_content_range(self):
+        a_response = _single_range_response
+        headers = http._extract_headers(a_response[1], 'http://nocontent')
+        del headers['Content-Range']
+        self.assertRaises(errors.InvalidHttpResponse,
+            response.handle_response, 'http://nocontent', a_response[0],
+                                      headers, StringIO(a_response[2]))

=== modified file 'bzrlib/tests/test_transport_implementations.py'
--- bzrlib/tests/test_transport_implementations.py	2006-07-03 19:27:26 +0000
+++ bzrlib/tests/test_transport_implementations.py	2006-07-14 19:25:24 +0000
@@ -709,8 +709,8 @@
         except (ConnectionError, NoSuchFile), e:
             pass
         except (Exception), e:
-            self.fail('Wrong exception thrown (%s): %s' 
-                        % (e.__class__.__name__, e))
+            self.fail('Wrong exception thrown (%s.%s): %s' 
+                        % (e.__class__.__module__, e.__class__.__name__, e))
         else:
             self.fail('Did not get the expected ConnectionError or NoSuchFile.')
 
@@ -988,10 +988,11 @@
         if transport.is_readonly():
             file('a', 'w').write('0123456789')
         else:
-            transport.put('a', StringIO('01234567890'))
+            transport.put('a', StringIO('0123456789'))
 
         d = list(transport.readv('a', ((0, 1), (1, 1), (3, 2), (9, 1))))
         self.assertEqual(d[0], (0, '0'))
         self.assertEqual(d[1], (1, '1'))
         self.assertEqual(d[2], (3, '34'))
         self.assertEqual(d[3], (9, '9'))
+

=== modified file 'bzrlib/transport/__init__.py'
--- bzrlib/transport/__init__.py	2006-07-07 21:40:36 +0000
+++ bzrlib/transport/__init__.py	2006-07-14 19:25:24 +0000
@@ -307,7 +307,7 @@
         
         Note that some transports MAY allow querying on directories, but this
         is not part of the protocol.  In other words, the results of 
-        t.has("a_directory_name") are undefined."
+        t.has("a_directory_name") are undefined.
         """
         raise NotImplementedError(self.has)
 
@@ -376,7 +376,7 @@
             if not combined_offsets:
                 combined_offsets = [[offset, size]]
             else:
-                if (len (combined_offsets) < 50 and
+                if (len(combined_offsets) < 50 and
                     combined_offsets[-1][0] + combined_offsets[-1][1] == offset):
                     # combatible offset:
                     combined_offsets.append([offset, size])

=== modified file 'bzrlib/transport/http/__init__.py'
--- bzrlib/transport/http/__init__.py	2006-06-11 03:40:17 +0000
+++ bzrlib/transport/http/__init__.py	2006-07-19 14:17:49 +0000
@@ -19,9 +19,9 @@
 There are separate implementation modules for each http client implementation.
 """
 
-from collections import deque
 from cStringIO import StringIO
 import errno
+import mimetools
 import os
 import posixpath
 import re
@@ -30,14 +30,18 @@
 import urllib
 from warnings import warn
 
-from bzrlib.transport import Transport, register_transport, Server
+# TODO: load these only when running http tests
+import BaseHTTPServer, SimpleHTTPServer, socket, time
+import threading
+
+from bzrlib import errors
 from bzrlib.errors import (TransportNotPossible, NoSuchFile,
                            TransportError, ConnectionError, InvalidURL)
 from bzrlib.branch import Branch
 from bzrlib.trace import mutter
-# TODO: load these only when running http tests
-import BaseHTTPServer, SimpleHTTPServer, socket, time
-import threading
+from bzrlib.transport import Transport, register_transport, Server
+from bzrlib.transport.http.response import (HttpMultipartRangeResponse,
+                                            HttpRangeResponse)
 from bzrlib.ui import ui_factory
 
 
@@ -71,6 +75,50 @@
     return url
 
 
+def _extract_headers(header_text, url):
+    """Extract the mapping for an rfc2822 header
+
+    This is a helper function for the test suite and for _pycurl.
+    (urllib already parses the headers for us)
+
+    In the case that there are multiple headers inside the file,
+    the last one is returned.
+
+    :param header_text: A string of header information.
+        This expects that the first line of a header will always be HTTP ...
+    :param url: The url we are parsing, so we can raise nice errors
+    :return: mimetools.Message object, which basically acts like a case 
+        insensitive dictionary.
+    """
+    first_header = True
+    remaining = header_text
+
+    if not remaining:
+        raise errors.InvalidHttpResponse(url, 'Empty headers')
+
+    while remaining:
+        header_file = StringIO(remaining)
+        first_line = header_file.readline()
+        if not first_line.startswith('HTTP'):
+            if first_header: # The first header *must* start with HTTP
+                raise errors.InvalidHttpResponse(url,
+                    'Opening header line did not start with HTTP: %s' 
+                    % (first_line,))
+                assert False, 'Opening header line was not HTTP'
+            else:
+                break # We are done parsing
+        first_header = False
+        m = mimetools.Message(header_file)
+
+        # mimetools.Message parses the first header up to a blank line
+        # So while there is remaining data, it probably means there is
+        # another header to be parsed.
+        # Get rid of any preceeding whitespace, which if it is all whitespace
+        # will get rid of everything.
+        remaining = header_file.read().lstrip()
+    return m
+
+
 class HttpTransportBase(Transport):
     """Base class for http implementations.
 
@@ -195,58 +243,43 @@
         :param offsets: A list of (offset, size) tuples.
         :param return: A list or generator of (offset, data) tuples
         """
-        # Ideally we would pass one big request asking for all the ranges in
-        # one go; however then the server will give a multipart mime response
-        # back, and we can't parse them yet.  So instead we just get one range
-        # per region, and try to coallesce the regions as much as possible.
-        #
-        # The read-coallescing code is not quite regular enough to have a
-        # single driver routine and
-        # helper method in Transport.
-        def do_combined_read(combined_offsets):
-            # read one coalesced block
-            total_size = 0
-            for offset, size in combined_offsets:
-                total_size += size
-            mutter('readv coalesced %d reads.', len(combined_offsets))
-            offset = combined_offsets[0][0]
-            byte_range = (offset, offset + total_size - 1)
-            code, result_file = self._get(relpath, [byte_range])
-            if code == 206:
-                for off, size in combined_offsets:
-                    result_bytes = result_file.read(size)
-                    assert len(result_bytes) == size
-                    yield off, result_bytes
-            elif code == 200:
-                data = result_file.read(offset + total_size)[offset:offset + total_size]
-                pos = 0
-                for offset, size in combined_offsets:
-                    yield offset, data[pos:pos + size]
-                    pos += size
-                del data
-        if not len(offsets):
-            return
-        pending_offsets = deque(offsets)
-        combined_offsets = []
-        while len(pending_offsets):
-            offset, size = pending_offsets.popleft()
-            if not combined_offsets:
-                combined_offsets = [[offset, size]]
+        ranges = self.offsets_to_ranges(offsets)
+        mutter('http readv of %s collapsed %s offsets => %s',
+                relpath, len(offsets), ranges)
+        code, f = self._get(relpath, ranges)
+        for start, size in offsets:
+            f.seek(start, (start < 0) and 2 or 0)
+            start = f.tell()
+            data = f.read(size)
+            assert len(data) == size
+            yield start, data
+
+    @staticmethod
+    def offsets_to_ranges(offsets):
+        """Turn a list of offsets and sizes into a list of byte ranges.
+
+        :param offsets: A list of tuples of (start, size).  An empty list
+            is not accepted.
+        :return: a list of inclusive byte ranges (start, end) 
+            Adjacent ranges will be combined.
+        """
+        # Make sure we process sorted offsets
+        offsets = sorted(offsets)
+
+        prev_end = None
+        combined = []
+
+        for start, size in offsets:
+            end = start + size - 1
+            if prev_end is None:
+                combined.append([start, end])
+            elif start <= prev_end + 1:
+                combined[-1][1] = end
             else:
-                if (len (combined_offsets) < 500 and
-                    combined_offsets[-1][0] + combined_offsets[-1][1] == offset):
-                    # combatible offset:
-                    combined_offsets.append([offset, size])
-                else:
-                    # incompatible, or over the threshold issue a read and yield
-                    pending_offsets.appendleft((offset, size))
-                    for result in do_combined_read(combined_offsets):
-                        yield result
-                    combined_offsets = []
-        # whatever is left is a single coalesced request
-        if len(combined_offsets):
-            for result in do_combined_read(combined_offsets):
-                yield result
+                combined.append([start, end])
+            prev_end = end
+
+        return combined
 
     def put(self, relpath, f, mode=None):
         """Copy the file-like or string object into the location.
@@ -343,6 +376,25 @@
         else:
             return self.__class__(self.abspath(offset))
 
+    @staticmethod
+    def range_header(ranges, tail_amount):
+        """Turn a list of bytes ranges into a HTTP Range header value.
+
+        :param offsets: A list of byte ranges, (start, end). An empty list
+        is not accepted.
+
+        :return: HTTP range header string.
+        """
+        strings = []
+        for start, end in ranges:
+            strings.append('%d-%d' % (start, end))
+
+        if tail_amount:
+            strings.append('-%d' % tail_amount)
+
+        return ','.join(strings)
+
+
 #---------------- test server facilities ----------------
 # TODO: load these only when running tests
 
@@ -435,6 +487,7 @@
                                                 RequestHandlerClass)
         self.test_case = test_case
 
+
 class HttpServer(Server):
     """A test server for http transports."""
 

=== modified file 'bzrlib/transport/http/_pycurl.py'
--- bzrlib/transport/http/_pycurl.py	2006-04-27 06:06:51 +0000
+++ bzrlib/transport/http/_pycurl.py	2006-07-19 14:17:49 +0000
@@ -24,15 +24,18 @@
 # It's probably safer to just always revalidate.  mbp 20060321
 
 import os
-from StringIO import StringIO
+from cStringIO import StringIO
 
+from bzrlib import errors
 import bzrlib
 from bzrlib.errors import (TransportNotPossible, NoSuchFile,
                            TransportError, ConnectionError,
                            DependencyNotPresent)
 from bzrlib.trace import mutter
 from bzrlib.transport import register_urlparse_netloc_protocol
-from bzrlib.transport.http import HttpTransportBase, extract_auth, HttpServer
+from bzrlib.transport.http import (HttpTransportBase, HttpServer,
+                                   _extract_headers,
+                                   response, _pycurl_errors)
 
 try:
     import pycurl
@@ -69,9 +72,15 @@
     set headers to allow caching.
     """
 
-    def __init__(self, base):
+    def __init__(self, base, from_transport=None):
         super(PyCurlTransport, self).__init__(base)
-        mutter('using pycurl %s' % pycurl.version)
+        if from_transport is not None:
+            self._base_curl = from_transport._base_curl
+            self._range_curl = from_transport._range_curl
+        else:
+            mutter('using pycurl %s' % pycurl.version)
+            self._base_curl = pycurl.Curl()
+            self._range_curl = pycurl.Curl()
 
     def should_cache(self):
         """Return True if the data pulled across should be cached locally.
@@ -79,12 +88,15 @@
         return True
 
     def has(self, relpath):
-        curl = pycurl.Curl()
+        """See Transport.has()"""
+        # We set NO BODY=0 in _get_full, so it should be safe
+        # to re-use the non-range curl object
+        curl = self._base_curl
         abspath = self._real_abspath(relpath)
         curl.setopt(pycurl.URL, abspath)
-        curl.setopt(pycurl.FOLLOWLOCATION, 1) # follow redirect responses
         self._set_curl_options(curl)
         # don't want the body - ie just do a HEAD request
+        # This means "NO BODY" not 'nobody'
         curl.setopt(pycurl.NOBODY, 1)
         self._curl_perform(curl)
         code = curl.getinfo(pycurl.HTTP_CODE)
@@ -92,38 +104,80 @@
             return False
         elif code in (200, 302): # "ok", "found"
             return True
-        elif code == 0:
-            self._raise_curl_connection_error(curl)
         else:
             self._raise_curl_http_error(curl)
         
-    def _get(self, relpath, ranges):
-        curl = pycurl.Curl()
+    def _get(self, relpath, ranges, tail_amount=0):
+        # This just switches based on the type of request
+        if ranges is not None or tail_amount not in (0, None):
+            return self._get_ranged(relpath, ranges, tail_amount=tail_amount)
+        else:
+            return self._get_full(relpath)
+    
+    def _setup_get_request(self, curl, relpath):
+        """Do the common setup stuff for making a request
+
+        :param curl: The curl object to place the request on
+        :param relpath: The relative path that we want to get
+        :return: (abspath, data, header) 
+                 abspath: full url
+                 data: file that will be filled with the body
+                 header: file that will be filled with the headers
+        """
         abspath = self._real_abspath(relpath)
-        sio = StringIO()
         curl.setopt(pycurl.URL, abspath)
         self._set_curl_options(curl)
-        curl.setopt(pycurl.WRITEFUNCTION, sio.write)
+        # Make sure we do a GET request. versions > 7.14.1 also set the
+        # NO BODY flag, but we'll do it ourselves in case it is an older
+        # pycurl version
         curl.setopt(pycurl.NOBODY, 0)
-        if ranges is not None:
-            assert len(ranges) == 1
-            # multiple ranges not supported yet because we can't decode the
-            # response
-            curl.setopt(pycurl.RANGE, '%d-%d' % ranges[0])
+        curl.setopt(pycurl.HTTPGET, 1)
+
+        data = StringIO()
+        header = StringIO()
+        curl.setopt(pycurl.WRITEFUNCTION, data.write)
+        curl.setopt(pycurl.HEADERFUNCTION, header.write)
+
+        return abspath, data, header
+
+    def _get_full(self, relpath):
+        """Make a request for the entire file"""
+        curl = self._base_curl
+        abspath, data, header = self._setup_get_request(curl, relpath)
         self._curl_perform(curl)
+
         code = curl.getinfo(pycurl.HTTP_CODE)
+        data.seek(0)
+
         if code == 404:
             raise NoSuchFile(abspath)
-        elif code == 200:
-            sio.seek(0)
-            return code, sio
-        elif code == 206 and (ranges is not None):
-            sio.seek(0)
-            return code, sio
-        elif code == 0:
-            self._raise_curl_connection_error(curl)
-        else:
-            self._raise_curl_http_error(curl)
+        if code != 200:
+            self._raise_curl_http_error(curl, 'expected 200 or 404 for full response.')
+
+        return code, data
+
+    def _get_ranged(self, relpath, ranges, tail_amount):
+        """Make a request for just part of the file."""
+        # We would like to re-use the same curl object for 
+        # full requests and partial requests
+        # Documentation says 'Pass in NULL to disable the use of ranges'
+        # None is the closest we have, but at least with pycurl 7.13.1
+        # It raises an 'invalid arguments' response
+        # curl.setopt(pycurl.RANGE, None)
+        # curl.unsetopt(pycurl.RANGE) doesn't support the RANGE parameter
+        # So instead we hack around this by using a separate objects
+        curl = self._range_curl
+        abspath, data, header = self._setup_get_request(curl, relpath)
+
+        curl.setopt(pycurl.RANGE, self.range_header(ranges, tail_amount))
+        self._curl_perform(curl)
+        data.seek(0)
+
+        code = curl.getinfo(pycurl.HTTP_CODE)
+        # mutter('url: %s header:\n%s', abspath, header.getvalue())
+        headers = _extract_headers(header.getvalue(), abspath)
+        # handle_response will raise NoSuchFile, etc based on the response code
+        return code, response.handle_response(abspath, code, headers, data)
 
     def _raise_curl_connection_error(self, curl):
         curl_errno = curl.getinfo(pycurl.OS_ERRNO)
@@ -131,11 +185,15 @@
         raise ConnectionError('curl connection error (%s) on %s'
                               % (os.strerror(curl_errno), url))
 
-    def _raise_curl_http_error(self, curl):
+    def _raise_curl_http_error(self, curl, info=None):
         code = curl.getinfo(pycurl.HTTP_CODE)
         url = curl.getinfo(pycurl.EFFECTIVE_URL)
-        raise TransportError('http error %d probing for %s' %
-                             (code, url))
+        if info is None:
+            msg = ''
+        else:
+            msg = ': ' + info
+        raise errors.InvalidHttpResponse(url, 'Unable to handle http code %d%s'
+                                              % (code,msg))
 
     def _set_curl_options(self, curl):
         """Set options for all requests"""
@@ -143,10 +201,11 @@
         # to force it to always retrieve.  so just turn off the default Pragma
         # provided by Curl.
         headers = ['Cache-control: max-age=0',
-                   'Pragma: no-cache']
+                   'Pragma: no-cache',
+                   'Connection: Keep-Alive']
         ## curl.setopt(pycurl.VERBOSE, 1)
         # TODO: maybe include a summary of the pycurl version
-        ua_str = 'bzr/%s (pycurl)' % (bzrlib.__version__)
+        ua_str = 'bzr/%s (pycurl)' % (bzrlib.__version__,)
         curl.setopt(pycurl.USERAGENT, ua_str)
         curl.setopt(pycurl.HTTPHEADER, headers)
         curl.setopt(pycurl.FOLLOWLOCATION, 1) # follow redirect responses
@@ -157,9 +216,15 @@
             curl.perform()
         except pycurl.error, e:
             # XXX: There seem to be no symbolic constants for these values.
-            if e[0] == 6:
-                # couldn't resolve host
-                raise NoSuchFile(curl.getinfo(pycurl.EFFECTIVE_URL), e)
+            url = curl.getinfo(pycurl.EFFECTIVE_URL)
+            mutter('got pycurl error: %s, %s, %s, url: %s ',
+                    e[0], _pycurl_errors.errorcode[e[0]], e, url)
+            if e[0] in (_pycurl_errors.CURLE_COULDNT_RESOLVE_HOST,
+                        _pycurl_errors.CURLE_COULDNT_CONNECT):
+                self._raise_curl_connection_error(curl)
+            # jam 20060713 The code didn't use to re-raise the exception here
+            # but that seemed bogus
+            raise
 
 
 class HttpServer_PyCurl(HttpServer):

=== added file 'bzrlib/transport/http/_pycurl_errors.py'
--- bzrlib/transport/http/_pycurl_errors.py	1970-01-01 00:00:00 +0000
+++ bzrlib/transport/http/_pycurl_errors.py	2006-07-14 01:46:36 +0000
@@ -0,0 +1,102 @@
+# Copyright (C) 2006 Canonical
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+# Taken from curl/curl.h
+CURLE_OK = 0
+CURLE_UNSUPPORTED_PROTOCOL = 1
+CURLE_FAILED_INIT = 2
+CURLE_URL_MALFORMAT = 3
+CURLE_URL_MALFORMAT_USER = 4 # (NOT USED)
+CURLE_COULDNT_RESOLVE_PROXY = 5
+CURLE_COULDNT_RESOLVE_HOST = 6
+CURLE_COULDNT_CONNECT = 7
+CURLE_FTP_WEIRD_SERVER_REPLY = 8
+CURLE_FTP_ACCESS_DENIED = 9 # a service was denied by the FTP server
+                            # due to lack of access - when login fails
+                            # this is not returned.
+CURLE_FTP_USER_PASSWORD_INCORRECT = 10
+CURLE_FTP_WEIRD_PASS_REPLY = 11
+CURLE_FTP_WEIRD_USER_REPLY = 12
+CURLE_FTP_WEIRD_PASV_REPLY = 13
+CURLE_FTP_WEIRD_227_FORMAT = 14
+CURLE_FTP_CANT_GET_HOST = 15
+CURLE_FTP_CANT_RECONNECT = 16
+CURLE_FTP_COULDNT_SET_BINARY = 17
+CURLE_PARTIAL_FILE = 18
+CURLE_FTP_COULDNT_RETR_FILE = 19
+CURLE_FTP_WRITE_ERROR = 20
+CURLE_FTP_QUOTE_ERROR = 21
+CURLE_HTTP_RETURNED_ERROR = 22
+CURLE_WRITE_ERROR = 23
+CURLE_MALFORMAT_USER = 24 # NOT USED
+CURLE_FTP_COULDNT_STOR_FILE = 25 # failed FTP upload
+CURLE_READ_ERROR = 26 # could open/read from file
+CURLE_OUT_OF_MEMORY = 27
+CURLE_OPERATION_TIMEOUTED = 28 # the timeout time was reached
+CURLE_FTP_COULDNT_SET_ASCII = 29 # TYPE A failed
+CURLE_FTP_PORT_FAILED = 30 # FTP PORT operation failed
+CURLE_FTP_COULDNT_USE_REST = 31 # the REST command failed
+CURLE_FTP_COULDNT_GET_SIZE = 32 # the SIZE command failed
+CURLE_HTTP_RANGE_ERROR = 33 # RANGE "command" didn't work
+CURLE_HTTP_POST_ERROR = 34
+CURLE_SSL_CONNECT_ERROR = 35 # wrong when connecting with SSL
+CURLE_BAD_DOWNLOAD_RESUME = 36 # couldn't resume download
+CURLE_FILE_COULDNT_READ_FILE = 37
+CURLE_LDAP_CANNOT_BIND = 38
+CURLE_LDAP_SEARCH_FAILED = 39
+CURLE_LIBRARY_NOT_FOUND = 40
+CURLE_FUNCTION_NOT_FOUND = 41
+CURLE_ABORTED_BY_CALLBACK = 42
+CURLE_BAD_FUNCTION_ARGUMENT = 43
+CURLE_BAD_CALLING_ORDER = 44 # NOT USED
+CURLE_INTERFACE_FAILED = 45 # CURLOPT_INTERFACE failed
+CURLE_BAD_PASSWORD_ENTERED = 46 # NOT USED
+CURLE_TOO_MANY_REDIRECTS  = 47 # catch endless re-direct loops
+CURLE_UNKNOWN_TELNET_OPTION = 48 # User specified an unknown option
+CURLE_TELNET_OPTION_SYNTAX  = 49 # Malformed telnet option
+CURLE_OBSOLETE = 50 # NOT USED
+CURLE_SSL_PEER_CERTIFICATE = 51 # peer's certificate wasn't ok
+CURLE_GOT_NOTHING = 52 # when this is a specific error
+CURLE_SSL_ENGINE_NOTFOUND = 53 # SSL crypto engine not found
+CURLE_SSL_ENGINE_SETFAILED = 54 # can not set SSL crypto engine as default
+CURLE_SEND_ERROR = 55 # failed sending network data
+CURLE_RECV_ERROR = 56 # failure in receiving network data
+CURLE_SHARE_IN_USE = 57 # share is in use
+CURLE_SSL_CERTPROBLEM = 58 # problem with the local certificate
+CURLE_SSL_CIPHER = 59 # couldn't use specified cipher
+CURLE_SSL_CACERT = 60 # problem with the CA cert (path?)
+CURLE_BAD_CONTENT_ENCODING = 61 # Unrecognized transfer encoding
+CURLE_LDAP_INVALID_URL = 62 # Invalid LDAP URL
+CURLE_FILESIZE_EXCEEDED = 63 # Maximum file size exceeded
+CURLE_FTP_SSL_FAILED = 64 # Requested FTP SSL level failed
+CURLE_SEND_FAIL_REWIND = 65 # Sending the data requires a rewind that failed
+CURLE_SSL_ENGINE_INITFAILED = 66 # failed to initialise ENGINE
+CURLE_LOGIN_DENIED = 67 # user, password or similar was not
+                        # accepted and we failed to login
+CURLE_TFTP_NOTFOUND = 68 # file not found on server
+CURLE_TFTP_PERM = 69 # permission problem on server
+CURLE_TFTP_DISKFULL = 70 # out of disk space on server
+CURLE_TFTP_ILLEGAL = 71 # Illegal TFTP operation
+CURLE_TFTP_UNKNOWNID = 72 # Unknown transfer ID
+CURLE_TFTP_EXISTS = 73 # File already exists
+CURLE_TFTP_NOSUCHUSER = 74 # No such user
+
+
+# Create the reverse mapping, so we can look things up.
+errorcode = {}
+for name, val in globals().items():
+    if name.startswith('CURLE'):
+        errorcode[val] = name

=== modified file 'bzrlib/transport/http/_urllib.py'
--- bzrlib/transport/http/_urllib.py	2006-06-20 03:57:11 +0000
+++ bzrlib/transport/http/_urllib.py	2006-07-14 04:58:29 +0000
@@ -16,15 +16,16 @@
 
 import errno
 import urllib, urllib2
+import errno
+from StringIO import StringIO
 
 import bzrlib  # for the version
-from bzrlib.errors import BzrError
+from bzrlib.errors import (TransportNotPossible, NoSuchFile, BzrError,
+                           TransportError, ConnectionError)
 from bzrlib.trace import mutter
 from bzrlib.transport import register_urlparse_netloc_protocol
-from bzrlib.transport.http import HttpTransportBase, extract_auth, HttpServer
-from bzrlib.errors import (TransportNotPossible, NoSuchFile,
-                           TransportError, ConnectionError)
-
+from bzrlib.transport.http import (HttpTransportBase, HttpServer,
+                                   extract_auth, response)
 
 register_urlparse_netloc_protocol('http+urllib')
 
@@ -42,21 +43,24 @@
 
 
 class HttpTransport_urllib(HttpTransportBase):
-    """Python urllib transport for http and https.
-    """
+    """Python urllib transport for http and https."""
 
     # TODO: Implement pipelined versions of all of the *_multi() functions.
 
-    def __init__(self, base):
+    def __init__(self, base, from_transport=None):
         """Set the base path where files will be stored."""
         super(HttpTransport_urllib, self).__init__(base)
+        # HttpTransport_urllib doesn't maintain any per-transport state yet
+        # so nothing to do with from_transport
 
-    def _get(self, relpath, ranges):
+    def _get(self, relpath, ranges, tail_amount=0):
         path = relpath
         try:
             path = self._real_abspath(relpath)
-            response = self._get_url_impl(path, method='GET', ranges=ranges)
-            return response.code, response
+            resp = self._get_url_impl(path, method='GET', ranges=ranges,
+                                      tail_amount=tail_amount)
+            return resp.code, response.handle_response(path,
+                                resp.code, resp.headers, resp)
         except urllib2.HTTPError, e:
             mutter('url error code: %s for has url: %r', e.code, path)
             if e.code == 404:
@@ -72,16 +76,11 @@
                              % (self.abspath(relpath), str(e)),
                              orig_error=e)
 
-    def _get_url_impl(self, url, method, ranges):
+    def _get_url_impl(self, url, method, ranges, tail_amount=0):
         """Actually pass get request into urllib
 
         :returns: urllib Response object
         """
-        if ranges:
-            range_string = ranges
-        else:
-            range_string = 'all'
-        mutter("get_url %s [%s]" % (url, range_string))
         manager = urllib2.HTTPPasswordMgrWithDefaultRealm()
         url = extract_auth(url, manager)
         auth_handler = urllib2.HTTPBasicAuthHandler(manager)
@@ -90,10 +89,11 @@
         request.method = method
         request.add_header('Pragma', 'no-cache')
         request.add_header('Cache-control', 'max-age=0')
-        request.add_header('User-Agent', 'bzr/%s (urllib)' % bzrlib.__version__)
-        if ranges:
-            assert len(ranges) == 1
-            request.add_header('Range', 'bytes=%d-%d' % ranges[0])
+        request.add_header('User-Agent',
+                           'bzr/%s (urllib)' % (bzrlib.__version__,))
+        if ranges or tail_amount:
+            bytes = 'bytes=' + self.range_header(ranges, tail_amount)
+            request.add_header('Range', bytes)
         response = opener.open(request)
         return response
 

=== added file 'bzrlib/transport/http/response.py'
--- bzrlib/transport/http/response.py	1970-01-01 00:00:00 +0000
+++ bzrlib/transport/http/response.py	2006-07-18 20:27:50 +0000
@@ -0,0 +1,279 @@
+# Copyright (C) 2006 Michael Ellerman
+#           modified by John Arbash Meinel (Canonical Ltd)
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Handlers for HTTP Responses.
+
+The purpose of these classes is to provide a uniform interface for clients
+to standard HTTP responses, single range responses and multipart range
+responses.
+"""
+
+
+from bisect import bisect
+from cStringIO import StringIO
+import re
+
+from bzrlib import errors
+from bzrlib.trace import mutter
+
+
+class ResponseRange(object):
+    """A range in a RangeFile-object."""
+
+    __slots__ = ['_ent_start', '_ent_end', '_data_start']
+
+    def __init__(self, ent_start, ent_end, data_start):
+        self._ent_start = ent_start
+        self._ent_end = ent_end
+        self._data_start = data_start
+
+    def __cmp__(self, other):
+        """Compare this to other.
+
+        We need this both for sorting, and so that we can
+        bisect the list of ranges.
+        """
+        if isinstance(other, int):
+            # Later on we bisect for a starting point
+            # so we allow comparing against a single integer
+            return cmp(self._ent_start, other)
+        else:
+            return cmp((self._ent_start, self._ent_end, self._data_start),
+                       (other._ent_start, other._ent_end, other._data_start))
+
+    def __str__(self):
+        return "%s(%s-%s,%s)" % (self.__class__.__name__,
+                                 self._ent_start, self._ent_end,
+                                 self._data_start)
+
+
+class RangeFile(object):
+    """File-like object that allow access to partial available data.
+
+    Specified by a set of ranges.
+    """
+
+    def __init__(self, path, input_file):
+        self._path = path
+        self._pos = 0
+        self._len = 0
+        self._ranges = []
+        self._data = input_file.read()
+
+    def _add_range(self, ent_start, ent_end, data_start):
+        """Add an entity range.
+
+        :param ent_start: Start offset of entity
+        :param ent_end: End offset of entity (inclusive)
+        :param data_start: Start offset of data in data stream.
+        """
+        self._ranges.append(ResponseRange(ent_start, ent_end, data_start))
+        self._len = max(self._len, ent_end)
+
+    def _finish_ranges(self):
+        self._ranges.sort()
+
+    def read(self, size):
+        """Read size bytes from the current position in the file.
+
+        Reading across ranges is not supported.
+        """
+        # find the last range which has a start <= pos
+        i = bisect(self._ranges, self._pos) - 1
+
+        if i < 0 or self._pos > self._ranges[i]._ent_end:
+            raise errors.InvalidRange(self._path, self._pos)
+
+        r = self._ranges[i]
+
+        # mutter('found range %s %s for pos %s', i, self._ranges[i], self._pos)
+
+        if (self._pos + size - 1) > r._ent_end:
+            raise errors.InvalidRange(self._path, self._pos)
+
+        start = r._data_start + (self._pos - r._ent_start)
+        end   = start + size
+        # mutter("range read %d bytes at %d == %d-%d", size, self._pos,
+        #         start, end)
+        self._pos += (end-start)
+        return self._data[start:end]
+
+    def seek(self, offset, whence=0):
+        if whence == 0:
+            self._pos = offset
+        elif whence == 1:
+            self._pos += offset
+        elif whence == 2:
+            self._pos = self._len + offset
+        else:
+            raise ValueError("Invalid value %s for whence." % whence)
+
+        if self._pos < 0:
+            self._pos = 0
+
+    def tell(self):
+        return self._pos
+
+
+class HttpRangeResponse(RangeFile):
+    """A single-range HTTP response."""
+
+    # TODO: jam 20060706 Consider compiling these regexes on demand
+    _CONTENT_RANGE_RE = re.compile(
+        '\s*([^\s]+)\s+([0-9]+)-([0-9]+)/([0-9]+)\s*$')
+
+    def __init__(self, path, content_range, input_file):
+        # mutter("parsing 206 non-multipart response for %s", path)
+        RangeFile.__init__(self, path, input_file)
+        start, end = self._parse_range(content_range, path)
+        self._add_range(start, end, 0)
+        self._finish_ranges()
+
+    @staticmethod
+    def _parse_range(range, path='<unknown>'):
+        """Parse an http Content-range header and return start + end
+
+        :param range: The value for Content-range
+        :param path: Provide to give better error messages.
+        :return: (start, end) A tuple of integers
+        """
+        match = HttpRangeResponse._CONTENT_RANGE_RE.match(range)
+        if not match:
+            raise errors.InvalidHttpRange(path, range,
+                                          "Invalid Content-range")
+
+        rtype, start, end, total = match.groups()
+
+        if rtype != 'bytes':
+            raise errors.InvalidHttpRange(path, range,
+                    "Unsupported range type '%s'" % (rtype,))
+
+        try:
+            start = int(start)
+            end = int(end)
+        except ValueError, e:
+            raise errors.InvalidHttpRange(path, range, str(e))
+
+        return start, end
+
+
+class HttpMultipartRangeResponse(RangeFile):
+    """A multi-range HTTP response."""
+    
+    _CONTENT_TYPE_RE = re.compile(
+        '^\s*multipart/byteranges\s*;\s*boundary\s*=\s*(.*?)\s*$')
+    
+    # Start with --<boundary>\r\n
+    # and ignore all headers ending in \r\n
+    # except for content-range:
+    # and find the two trailing \r\n separators
+    # indicating the start of the text
+    # TODO: jam 20060706 This requires exact conformance
+    #       to the spec, we probably could relax the requirement
+    #       of \r\n, and use something more like (\r?\n)
+    _BOUNDARY_PATT = (
+        "^--%s(?:\r\n(?:(?:content-range:([^\r]+))|[^\r]+))+\r\n\r\n")
+
+    def __init__(self, path, content_type, input_file):
+        # mutter("parsing 206 multipart response for %s", path)
+        # TODO: jam 20060706 Is it valid to initialize a
+        #       grandparent without initializing parent?
+        RangeFile.__init__(self, path, input_file)
+
+        self.boundary_regex = self._parse_boundary(content_type, path)
+
+        for match in self.boundary_regex.finditer(self._data):
+            ent_start, ent_end = HttpRangeResponse._parse_range(match.group(1),
+                                                                path)
+            self._add_range(ent_start, ent_end, match.end())
+
+        self._finish_ranges()
+
+    @staticmethod
+    def _parse_boundary(ctype, path='<unknown>'):
+        """Parse the Content-type field.
+        
+        This expects a multipart Content-type, and returns a
+        regex which is capable of finding the boundaries
+        in the multipart data.
+        """
+        match = HttpMultipartRangeResponse._CONTENT_TYPE_RE.match(ctype)
+        if not match:
+            raise errors.InvalidHttpContentType(path, ctype,
+                    "Expected multipart/byteranges with boundary")
+
+        boundary = match.group(1)
+        # mutter('multipart boundary is %s', boundary)
+        pattern = HttpMultipartRangeResponse._BOUNDARY_PATT
+        return re.compile(pattern % re.escape(boundary),
+                          re.IGNORECASE | re.MULTILINE)
+
+
+def _is_multipart(content_type):
+    return content_type.startswith('multipart/byteranges;')
+
+
+def handle_response(url, code, headers, data):
+    """Interpret the code & headers and return a HTTP response.
+
+    This is a factory method which returns an appropriate HTTP response
+    based on the code & headers it's given.
+
+    :param url: The url being processed. Mostly for error reporting
+    :param code: The integer HTTP response code
+    :param headers: A dict-like object that contains the HTTP response headers
+    :param data: A file-like object that can be read() to get the
+                 requested data
+    :return: A file-like object that can seek()+read() the 
+             ranges indicated by the headers.
+    """
+
+    if code == 206:
+        try:
+            content_type = headers['Content-Type']
+        except KeyError:
+            raise errors.InvalidHttpContentType(url, '',
+                msg='Missing Content-Type')
+
+        if _is_multipart(content_type):
+            # Full fledged multipart response
+            return HttpMultipartRangeResponse(url, content_type, data)
+        else:
+            # A response to a range request, but not multipart
+            try:
+                content_range = headers['Content-Range']
+            except KeyError:
+                raise errors.InvalidHttpResponse(url,
+                    'Missing the Content-Range header in a 206 range response')
+            return HttpRangeResponse(url, content_range, data)
+    elif code == 200:
+        # A regular non-range response, unfortunately the result from
+        # urllib doesn't support seek, so we wrap it in a StringIO
+        tell = getattr(data, 'tell', None)
+        if tell is None:
+            return StringIO(data.read())
+        return data
+    elif code == 404:
+        raise errors.NoSuchFile(url)
+
+    # TODO: jam 20060713 Properly handle redirects (302 Found, etc)
+    #       The '_get' code says to follow redirects, we probably 
+    #       should actually handle the return values
+    else:
+        raise errors.InvalidHttpResponse(url, "Unknown response code %s" 
+                                              % (code,))
+

